# wiimote data analysis/filtering abstractions
#
# acc_filter (used for calculating up to 8 different hits)
# joy_filter (convert analog joystick into digital 9-switch controller)
# xyzaccel (convert xyz accelerometer data into rms-friendly format)
# accel_smoother (filter any data using median)
#
# Copyright (c) 2009 Ivica Ico Bukvic (ico@vt.edu)


try:
	import pyext
except:
	print "ERROR: This script must be loaded by the PD/Max pyext external"

#################################################################

# helper function - determine whether argument is a numeric type
def isNumber(value):
	import types
	if type(value) in (types.FloatType, types.IntType, types.LongType):
		return 1
	else:
		return 0

class acc_filter(pyext._class):
	"""Filter out Wiimote accelerometer data"""

	# number of inlets and outlets
	# inlets: 1 = delta, 2 = pitch, 3 = roll, 4 = x, 5 = y, 6 = z
	# outlets: 1 = value over threshold, 2 = median pitch, 3 = median roll,
	#		   4 = x delta, 5 = y delta, 6 = z delta
	#_inlets=6
	#_outlets=6

	# Constructor
	def __init__(self,*args):
		#accelerometer info
		self.reset_counter = 0
		self.reset_repeat = 10
		self.u_thresh = 0.4
		self.d_thresh = 0.05
		self.avg_thresh = 0.9

		self.delta = 0

		self.x_array = [0]
		self.y_array = [0]
		self.z_array = [0]
		self.xyz_size = 10

		self.pitch_array = [0]
		self.pitch_filter = [0]
		self.roll_array = [0]
		self.roll_filter = [0]

		self.a_size = 20

	# methods for first inlet

	def toggle_1(self, i):
		if (i == 1):
			tmp = self.pitch_array[len(self.pitch_array)-1]
			del self.pitch_array
			self.pitch_array = [tmp]

			tmp = self.roll_array[len(self.roll_array)-1]
			del self.roll_array
			self.roll_array = [tmp]

	def float_1(self,f):

		if (f > self.u_thresh):

			self.delta = f

			if (self.reset_counter == 0):

				self._outlet(6, sum(self.z_array)/len(self.z_array))
				self._outlet(5, sum(self.y_array)/len(self.y_array))
				self._outlet(4, sum(self.x_array)/len(self.x_array))

				tmp = sorted(self.roll_array)
				median = tmp[int(len(tmp)/2)]
				self._outlet(3, median)

				tmp = sorted(self.pitch_array)
				median = tmp[int(len(tmp)/2)]
				self._outlet(2, median)

				self._outlet(1, f)

			self.reset_counter = self.reset_repeat

		if (f < self.d_thresh) and (self.reset_counter > 0):
			self.reset_counter -= 1

	def float_2(self,f):
		tmp = sorted(self.pitch_filter)
		median = tmp[int(len(tmp)/2)]
		if (abs(f-median) < self.avg_thresh):
			self.pitch_array.append(round(f,0))
			self.pitch_filter.append(round(f,0))
			if (len(self.pitch_array) > self.a_size):
				del self.pitch_array[0]
			if (len(self.pitch_filter) > self.a_size):
				del self.pitch_filter[0]
		else:
			self.pitch_filter.append(round(f,0))
			if (len(self.pitch_filter) > self.a_size):
				del self.pitch_filter[0]		

	def float_3(self,f):
		tmp = sorted(self.roll_filter)
		median = tmp[int(len(tmp)/2)]
		if (abs(f-median) < self.avg_thresh):
			self.roll_array.append(round(f,0))
			self.roll_filter.append(round(f,0))
			if (len(self.roll_array) > self.a_size):
				del self.roll_array[0]
			if (len(self.roll_filter) > self.a_size):
				del self.roll_filter[0]
		else:
			self.roll_filter.append(round(f,0))
			if (len(self.roll_filter) > self.a_size):
				del self.roll_filter[0]	

	def float_4(self,f):
		self.x_array.append(f)
		if (len(self.x_array) > self.xyz_size):
			del self.x_array[0]

	def float_5(self,f):
		self.y_array.append(f)
		if (len(self.y_array) > self.xyz_size):
			del self.y_array[0]

	def float_6(self,f):
		self.z_array.append(f)
		if (len(self.z_array) > self.xyz_size):
			del self.z_array[0]

	# methods for control data

	def uthresh_1(self, f):
		if isNumber(f):
			self.u_thresh = f
			print "u_thresh = ", self.u_thresh
		else:
			print "acc_filter: ignoring uthresh garbage"		

	def dthresh_1(self, f):
		if isNumber(f):
			self.d_thresh = f
			print "d_thresh = ", self.d_thresh
		else:
			print "acc_filter: ignoring dthresh garbage"

	def avgthresh_1(self, f):
		if isNumber(f):
			self.avg_thresh = f
			print "avg_thresh = ", self.avg_thresh
		else:
			print "acc_filter: ignoring avgthresh garbage"

	def debug_1(self):
		tmp = sorted(self.x_array)
		median = tmp[int(len(tmp)/2)]
		print "median = ", median

#################################################################

class joy_filter(pyext._class):
	"""Filter out Joystick analog data"""

	# x and y
	#_inlets = 2
	# 1-8 directions starting with 1 being left and going counter-clockwise
	#_outlets = 1

	def __init__(self,*args):
		self.x = 130
		self.y = 130

		self.output = 0

	def float_1(self,f):
		self.x = f

		#do calculations
		if (self.x > 100) and (self.x < 160):
			#x axis is at 0
			if (self.y > 100) and (self.y < 160):
				#no change
				if (self.output != 0):
					self.output = 0
					self._outlet(1, self.output)
			elif (self.y > 210):
				#up
				if (self.output != 7):
					self.output = 7
					self._outlet(1, self.output)
			elif (self.y < 50):
				#down
				if (self.output != 3):
					self.output = 3
					self._outlet(1, self.output)
		elif (self.y > 100) and (self.y < 160):
			#y axis is at 0
			if (self.x > 210):
				#right
				if (self.output != 5):
					self.output = 5
					self._outlet(1, self.output)
			elif (self.x < 50):
				#left
				if (self.output != 1):
					self.output = 1
					self._outlet(1, self.output)

		#otherwise we are dealing with diagonal versions
		elif (self.x > 190) and (self.y > 190):
			#up-right
			if (self.output != 6):
				self.output = 6
				self._outlet(1, self.output)				 
		elif (self.x > 190) and (self.y < 70):
			#down-right
			if (self.output != 4):
				self.output = 4
				self._outlet(1, self.output)				
		elif (self.x < 70) and (self.y > 190):
			#up-left
			if (self.output != 8):
				self.output = 8
				self._outlet(1, self.output)	
		elif (self.x < 70) and (self.y < 70):
			#down-left
			if (self.output != 2):
				self.output = 2
				self._outlet(1, self.output)

	def float_2(self,f):
		self.y = f

#################################################################

class xyzaccel2rms(pyext._class):
	"""Filter out xyz accelerometer delta data for rms"""

	# x y z
	#_inlets = 3
	# cooked and default delta
	#_outlets = 2

	def __init__(self,*args):
		self.x = 0
		self.y = 0
		self.z = 0

		self.output_array = [0]
		self.a_size = 10

		self.div = 0.3
		self.pow = 1.2
		self.limit = 1.0
		self.offset = 0.01

		self.prev_delta = 0
		self.old_outcome = 0

	def float_1(self,f):
		self.x = f

		#get the delta
		cur_delta = pow(pow(self.x,2)+pow(self.y,2)+pow(self.z,2),0.5)
		self.output_array.append(abs(cur_delta-self.prev_delta))
		self.prev_delta = cur_delta
		if (len(self.output_array) > self.a_size):
			del self.output_array[0]

		#calculate average
		avg = sum(self.output_array)/len(self.output_array)

		#scaling
		avg = pow(avg, self.pow)
		if (avg > self.limit):
			avg = self.limit

		#cook the output
		if (avg < 0.01):
			outcome = round((avg/self.div),3)
		elif (avg < 0.1):
			outcome = round((avg/self.div),2)
		else:
			outcome = round((avg/self.div),1)

		outcome = outcome - self.offset
		if (outcome < 0):
			outcome = 0

		if (self.output_array[len(self.output_array)-1] < 0.01):
			uncooked = round(self.output_array[len(self.output_array)-1],3)
		elif (self.output_array[len(self.output_array)-1] < 0.1):
			uncooked = round(self.output_array[len(self.output_array)-1],2)
		else:
			uncooked = round(self.output_array[len(self.output_array)-1],1)

		if (uncooked > self.limit):
			uncooked = self.limit

		#print "avg = ", avg, " outcome = ", outcome

		#output
		self._outlet(2, uncooked)
		#print "cur_delta = ", self.output_array[len(self.output_array)-1]

		if (outcome != self.old_outcome):
			self._outlet(1, outcome)
			self.old_outcome = outcome

	def float_2(self,f):
		self.y = f

	def float_3(self,f):
		self.z = f

	def divide_1(self, f):
		if isNumber(f):
			self.div = f
			print "divide = ", self.div
		else:
			print "xyzaccel2rms: ignoring divide garbage"

	def limit_1(self, f):
		if isNumber(f) and (f > 0):
			self.limit = f
			print "limit = ", self.limit
		else:
			print "xyzaccel2rms: ignoring limit garbage"

	def pow_1(self, f):
		if isNumber(f) and (f > 0):
			self.pow = f
			print "pow = ", self.pow
		else:
			print "xyzaccel2rms: ignoring pow garbage"

	def offset_1(self, f):
		if isNumber(f) and (f > 0):
			self.offset = f
			print "offset = ", self.offset
		else:
			print "xyzaccel2rms: ignoring offset garbage"

	def bang_1(self):
		del self.output_array
		self.output_array = [0]

#################################################################

class accel_smoother(pyext._class):
	"""Filter any stream of data (e.g. accelerometer) using median"""

	# accel
	#_inlets = 1
	# cooked data
	#_outlets = 1

	# Constructor
	def __init__(self,*args):
		#accelerometer info
		self.u_thresh = 0.05
		self.old_x = 0
		self.x_array = [0]
		self.a_size = 20

	def float_1(self,f):

		if (abs(f-self.old_x) < self.u_thresh):

			self.x_array.append(round(f,2))
			if (len(self.x_array) > self.a_size):
				del self.x_array[0]

			tmp = sorted(self.x_array)
			median = tmp[int(len(tmp)/2)]
			self._outlet(1, median)

		self.old_x = f

	def uthresh_1(self, f):
		if isNumber(f):
			self.u_thresh = f
			print "u_thresh = ", self.u_thresh
		else:
			print "accel_smoother: ignoring uthresh garbage"	

	def asize_1(self, f):
		if isNumber(f):
			self.a_size = f
			print "a_size = ", self.a_size
		else:
			print "accel_smoother: ignoring a_size garbage"	

#################################################################
